"""
Service Plans

This module contains actions for viewing & selecting subscription Service Plans.

!!! question "Minimum Plugin Version: 5.0.0"
"""
import logging

from typing import List, Union, Optional, Dict

from playbook_tools import tables

from . import _client, users, subscriptions

from .records import ServicePlan, GraphClient


logger = logging.getLogger(__name__)


TABLE_HEADER = ["Name", "ID", "Applies To"]


def _row_generator(service_plan):
    return [service_plan["name"].replace("_", " "), service_plan["id"], service_plan.get("applies_to")] if service_plan else []


@_client.graph_api
def get_enabled_for_user(
    client: GraphClient, user_principal_name: str, part_number: str, user_only: Optional[bool] = False
) -> List[ServicePlan]:
    """Get a user's **enabled** service plans for a given subscription.

    !!! question "Minimum Plugin Version: 5.0.0"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

        - __user_principal_name:__ the [User's](../inputs_outputs/#user) principal name

        - __part_number:__ the [Subscription](../inputs_outputs/#subscription) part number

        - __user_only:__ set to `true` to only get plans that apply to user objects (defaults to `false`)

    ??? success "Output"

         A list of [ServicePlans](../inputs_outputs/#serviceplan).

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.service_plans.get_enabled_for_user:
            part_number: O365_BUSINESS_PREMIUM
            user_principal_name: example.user@domain.net
          load:
            client: graph_secrets
          save: enabled_service_plans
        ```
    """
    subscription = subscriptions.get_by_part_number(client, part_number)

    logger.info(f"getting enabled service plans for user {user_principal_name} & subscription {part_number}")
    with _client._exception_handler(f"get enabled service plans for user '{user_principal_name}'"):

        # Get all assigned licences for this user, but specifically ask for service plan information in the response.
        assignments = _client.get(
            client,
            (
                f"/users/{users.get_by_principal_name(client, user_principal_name)['id']}"
                "?$select=assignedLicenses,assignedPlans"
            ),
        )

    # It's possible the user hasn't been assigned the supplied licence, so check first.
    try:
        assignment = next(iter([lic for lic in assignments["assignedLicenses"] if lic["skuId"] == subscription["sku_id"]]))
    except StopIteration:
        logger.info(f"user {user_principal_name} has not been assigned subscription {part_number}")
        return []

    # No disabled plans, return all service plans for the subscription.
    if not assignment["disabledPlans"]:
        if user_only:
            return list(filter(lambda plan: plan["applies_to"].lower() == "user", subscription["service_plans"]))
        return subscription["service_plans"]

    # Now return every service plan for this subscription that *isn't* in the user's
    # disabledPlans. These will be the enabled plans for the supplied subscription.
    enabled_plans = list(filter(lambda plan: plan["id"] not in assignment["disabledPlans"], subscription["service_plans"]))
    if user_only:
        return list(filter(lambda plan: plan["applies_to"].lower() == "user", enabled_plans))
    return enabled_plans


@_client.graph_api
def get_disabled_for_user(
    client: GraphClient, user_principal_name: str, part_number: str, user_only: Optional[bool] = False
) -> List[ServicePlan]:
    """Get a user's **disabled** service plans for a given subscription.

    !!! question "Minimum Plugin Version: 5.0.0"

    ??? info "Input"

        - __client:__ a [GraphClient](../inputs_outputs/#graphclient)

        - __user_principal_name:__ the [User's](../inputs_outputs/#user) principal name

        - __part_number:__ the [Subscription](../inputs_outputs/#subscription) part number

        - __user_only:__ set to `true` to only get plans that apply to user objects (defaults to `false`)

    ??? success "Output"

         A list of [ServicePlans](../inputs_outputs/#serviceplan).

    ??? example
        ```YAML linenums="1"
        - microsoft_graph.service_plans.get_disabled_for_user:
            part_number: O365_BUSINESS_PREMIUM
            user_principal_name: example.user@domain.net
          load:
            client: graph_secrets
          save: disabled_service_plans
        ```
    """
    subscription = subscriptions.get_by_part_number(client, part_number)

    logger.info(f"getting disabled service plans for user {user_principal_name} & subscription {part_number}")
    with _client._exception_handler(f"get disabled service plans for user '{user_principal_name}'"):

        # Get all assigned licences for this user, but specifically ask for service plan information in the response.
        assignments = _client.get(
            client,
            (
                f"/users/{users.get_by_principal_name(client, user_principal_name)['id']}"
                "?$select=assignedLicenses,assignedPlans"
            ),
        )

    # It's possible the user hasn't been assigned the supplied licence, so check first.
    try:
        assignment = next(iter([lic for lic in assignments["assignedLicenses"] if lic["skuId"] == subscription["sku_id"]]))
    except StopIteration:
        logger.info(f"user {user_principal_name} has not been assigned subscription {part_number}")
        return []

    # No disabled plans, return an empty list.
    if not assignment["disabledPlans"]:
        return []

    # Now return every service plan for this subscription that *is* in the user's
    # disabledPlans. These will be the disabled plans for the supplied subscription.
    disabled_plans = list(filter(lambda plan: plan["id"] in assignment["disabledPlans"], subscription["service_plans"]))
    if user_only:
        return list(filter(lambda plan: plan["applies_to"].lower() == "user", disabled_plans))
    return disabled_plans


def display(text: str, service_plans: Union[List[ServicePlan], ServicePlan]) -> None:
    """Display service plans in a table.

The table will have the following columns:

- Name
- ID
- Applies To

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __service_plans:__ any number of [ServicePlans](../inputs_outputs/#serviceplan)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - microsoft_graph.subscriptions.get_by_part_number:
        part_number: cbdc14ab-d96c-4c30-b9f4-6ada7cdc1d46
      load:
        client: graph_secrets
      save: subscription

    - microsoft_graph.service_plans.display:
        text: Subscription Service Plans
      load:
        service_plans: subscription.service_plans
    ```
    """
    tables.output_table(text, header=TABLE_HEADER, extract=_row_generator, records=service_plans)


def display_tabbed(text: str, tabs: Dict[str, Union[ServicePlan, List[ServicePlan]]]) -> None:
    """Display service plans in a table with multiple tabs.

The table will have the following columns:

- Name
- ID
- Applies To

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is any number of [ServicePlans](../inputs_outputs/#serviceplan)

??? success "Output"

    Nothing is outputted by this action.

??? example

    Auditing a user's enabled & disabled service plans for a given licence & displaying them in a tabbed table:

    ```YAML linenums="1"
    - microsoft_graph.service_plans.get_enabled_for_user:
        user_principal_name: example.user@domain.net
        part_number: cbdc14ab-d96c-4c30-b9f4-6ada7cdc1d46
      load:
        client: graph_secret
      save: enabled_plans

    - microsoft_graph.service_plans.get_disabled_for_user:
        user_principal_name: example.user@domain.net
        part_number: cbdc14ab-d96c-4c30-b9f4-6ada7cdc1d46
      load:
        client: graph_secret
      save: disabled_plans

    - ppa_tools.dictionaries.create:
      load:
        Enabled: enabled_plans
        Disabled: disabled_plans
      save: table_tabs

    - microsoft_graph.service_plans.display_tabbed:
        text: Enabled & Disabled Service Plans
      load:
        tabs: table_tabs
    ```
    """
    tables.output_tabbed_table(text, header=TABLE_HEADER, extract=_row_generator, tabs=tabs)


def select(
    text: str, service_plans: Union[List[ServicePlan], ServicePlan], minimum: Optional[int] = None, maximum: Optional[int] = None,
) -> None:
    """Display service plans in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- ID
- Applies To

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __service_plans:__ any number of [ServicePlans](../inputs_outputs/#serviceplan)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [ServicePlans](../inputs_outputs/#serviceplan)

    - **all**: a list of selected [ServicePlans](../inputs_outputs/#serviceplan)

    - **first**: the first selected [ServicePlan](../inputs_outputs/#serviceplan)

??? example
    ```YAML linenums="1"
    - microsoft_graph.subscriptions.get_by_part_number:
        part_number: cbdc14ab-d96c-4c30-b9f4-6ada7cdc1d46
      load:
        client: graph_secrets
      save: subscription

    - microsoft_graph.service_plans.select:
        text: Select Service Plans
      load:
        service_plans: subscription.service_plans
      save: selection
    ```
    """
    return tables.input_table(
        text, header=TABLE_HEADER, extract=_row_generator, records=service_plans, minimum=minimum, maximum=maximum
    )


def select_one(text: str, service_plans: Union[List[ServicePlan], ServicePlan]) -> None:
    """Display service plans in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- ID
- Applies To

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __service_plans:__ any number of [ServicePlans](../inputs_outputs/#serviceplan)

??? success "Output"

    A single [ServicePlan](../inputs_outputs/#serviceplan).

??? example

    ```YAML linenums="1"
    - microsoft_graph.subscriptions.get_by_part_number:
        part_number: cbdc14ab-d96c-4c30-b9f4-6ada7cdc1d46
      load:
        client: graph_secrets
      save: subscription

    - microsoft_graph.service_plans.select_one:
        text: Select Service Plan
      load:
        service_plans: subscription.service_plans
      save: service_plan
    ```
    """
    return select(text, service_plans=service_plans, minimum=1, maximum=1)["first"]
